﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Management;
using System.Text.RegularExpressions;
using WebConfigEncrypt.Types;
using System.IO;
using System.Xml.Linq;
using System.Threading;
using WebConfigEncrypt.Properties;

namespace WebConfigEncrypt.Interfaces {
    
    /// <summary>
    /// Performs WMI work to connect to a server and make changes
    /// to the IIS settings
    /// </summary>
    public class WebServerInterface {

        #region Constants

        private const string REGEX_DIRECTORY_INFORMATION = @"w3svc/(?<id>\d+)/root(?<path>.*)$";
        private const string WEBCONFIG_SHARE_NAME = "IIITempWebConfigEncryptShareIII";
        private const string DEFAULT_SITE_NAME = "<Unknown Name>";

        #endregion

        #region Properties

        /// <summary>
        /// The username to use to talk to the server with
        /// </summary>
        public string Username { get; set; }

        /// <summary>
        /// The password to use for the username
        /// </summary>
        public string Password { get; set; }

        /// <summary>
        /// The target server for these requests
        /// </summary>
        public string Server { get; set; }

        /// <summary>
        /// The domain being used
        /// </summary>
        public string Domain { get; set; }

        #endregion

        #region Working With The Server

        /// <summary>
        /// Returns a list of the ServerComment name for each site (somthing that 
        /// a person is more likely to recognize)
        /// </summary>
        public string GetSiteName(string name) {

            //get the scope
            ManagementScope scope = null;
            try {
                scope = this._CreateWmiScope(@"root\MicrosoftIISv2");
                scope.Connect();
            }
            catch {
                return name;
            }

            //if this doesn't work we aren't that concerned because they can
            //just go by the actual site id/name
            ManagementObjectSearcher setting = new ManagementObjectSearcher(scope,
                new ObjectQuery(string.Format("select serverComment from IIsWebServerSetting where Name = '{0}'", name)));
            try {
                foreach (ManagementObject result in setting.Get()) {
                    return result["serverComment"] as string;
                }
            }
            catch (Exception) {
                return name;
            }
            finally {
                setting.Dispose();
            }

            //nothing was found
            return name;

        }

        /// <summary>
        /// Returns a list of all virtual directories
        /// </summary>
        public IEnumerable<IISDirectory> GetVirtualDirectories() {

            //get the scope
            ManagementScope scope = null;
            try {
                scope = this._CreateWmiScope(@"root\MicrosoftIISv2");
                scope.Connect();
            }
            catch {
                throw new InvalidOperationException();
            }

            //get the results
            ManagementObjectSearcher search = null;
            ManagementObjectCollection results = null;
            try {
                //create the query
                search = new ManagementObjectSearcher(scope, new ObjectQuery("select Name, Path from IISWebVirtualDirSetting"));
                results = search.Get();
            }
            catch {
                throw new ManagementException();
            }
            finally {
                search.Dispose();
            }

            //get each virtual directory
            foreach (ManagementObject item in results) {

                //get the information about this site
                string name = (item["name"] as string ?? string.Empty).Trim();
                name = string.IsNullOrEmpty(name) ? "/" : name;

                //get the details of the id
                Match detail = Regex.Match(name, WebServerInterface.REGEX_DIRECTORY_INFORMATION, RegexOptions.IgnoreCase);
                string site = Regex.Match(name, @"(?<site>^w3svc/\d+)", RegexOptions.IgnoreCase).Groups["site"].Value;

                //get other information
                string id = (detail.Groups["id"].Value ?? string.Empty).Trim();
                string path = (item["path"] as string ?? string.Empty).Trim();
                string virt = (detail.Groups["path"].Value ?? string.Empty).Trim();
                virt = string.IsNullOrEmpty(virt) ? "/" : virt;

                //return this instance
                yield return new IISDirectory() {
                    Name = name,
                    Site = site,
                    Identity = id,
                    VirtualPath = virt,
                    PhysicalPath = path
                };

            }


        }

        /// <summary>
        /// Returns all web.config files from a directory
        /// </summary>
        public IEnumerable<WebConfigInformation> FindWebConfigFiles(IISDirectory directory) {

            //get the scope
            ManagementScope scope = null;
            try {
                scope = this._CreateWmiScope(@"root\CIMV2");
                scope.Connect();
            }
            catch {
                throw new InvalidOperationException();
            }

            //get the information
            string drive = Path.GetPathRoot(directory.PhysicalPath);
            string path = directory.PhysicalPath.Substring(drive.Length - 1);
            path = path.EndsWith("\\") ? path : string.Concat(path, "\\");

            //generate the query string
            string query = Regex.Replace(string.Format(
                "select Name from CIM_DataFile where Drive = '{0}' and Path like '{1}%' and FileName = 'web' and Extension = 'config'",
                Regex.Replace(drive, @"\\", string.Empty),
                path
                ), @"\\", @"\\");

            //query for the virtual directories
            ManagementObjectSearcher search = null;
            ManagementObjectCollection results = null;
            try {
                search = new ManagementObjectSearcher(scope, new ObjectQuery(query));
                results = search.Get();
            }
            catch {
                throw new ManagementException();
            }
            finally {
                search.Dispose();
            }

            //get each virtual directory
            foreach (var item in search.Get()) {
                yield return new WebConfigInformation() {
                    Path = item["Name"].ToString(),
                    Directory = directory
                };
            }

        }

        /// <summary>
        /// Executes a command on the remote server
        /// </summary>
        public bool ExecuteRemoteCommand(string command) {

            //get the scope
            ManagementScope scope = null;
            try {
                scope = this._CreateWmiScope(@"root\CIMV2");
                scope.Connect();
            }
            catch {
                throw new InvalidOperationException();
            }

            //Access the WMI class to invoke the command
            ManagementClass instance = new ManagementClass(scope, new ManagementPath("Win32_Process"), null);
            try {
                object result = instance.InvokeMethod(
                    "Create",
                    new object[] { command }
                    );
                //make sure the call was successful
                return result is int ? ((int)result == 0) : false;
            }
            catch {
                throw new ManagementException();
            }
            finally {
                instance.Dispose();
            }

        }

        /// <summary>
        /// Tries to read a file from a remote share
        /// </summary>
        public XDocument ReadRemoteFile(string file) {

            //because the file share may not be available, wait a couple seconds            
            int safety = 30;
            string result = null;
            while (safety-- > 0) {

                //wait before attempting to read - I HATE THIS!!
                Thread.Sleep(500);
                try {
                    if (File.Exists(file)) {
                        result = File.ReadAllText(file);
                        break;
                    }
                }
                catch {
                    //holy cow I hate doing this, but aspnet_regiis updates the time
                    //on the file before the update even starts so last changed time is
                    //unreliable. Additionally, while the file is being changed, trying
                    //to read it throws an IOException so that is unreliable as well.
                    //so this keeps polling the file until it is read
                }

            }

            //handle returning a document
            if (string.IsNullOrEmpty(result)) {
                System.Windows.Forms.MessageBox.Show("Couldn't ever read the file.");
                return null; 
            }
            try {
                return XDocument.Parse(result);
            }
            catch {
                return null;
            }
        }

        /// <summary>
        /// Sends a request to encrypt a section in a Web.config file
        /// </summary>
        public void EncryptSection(WebConfigInformation info, string section) {
            this.ExecuteRemoteCommand(
                string.Format(
                @"C:\WINDOWS\Microsoft.NET\Framework\v2.0.50727\aspnet_regiis -pe ""{0}"" -site ""{1}"" -app ""{2}""",
                section,
                info.Directory.Identity,
                info.Directory.VirtualPath
                ));
        }

        /// <summary>
        /// Sends a request to decrypt a section in a Web.config file
        /// </summary>
        public void DecryptSection(WebConfigInformation info, string section) {
            this.ExecuteRemoteCommand(
                string.Format(
                @"C:\WINDOWS\Microsoft.NET\Framework\v2.0.50727\aspnet_regiis -pd ""{0}"" -site ""{1}"" -app ""{2}""",
                section,
                info.Directory.Identity,
                info.Directory.VirtualPath
                ));
        }

        #endregion

        #region Accessing WMI Scope

        //prepares the scope for use
        private ManagementScope _CreateWmiScope(string path) {

            //prepare the connection
            ConnectionOptions connect = new ConnectionOptions();

            //if this is a remote request, provide credentials
            if (!Regex.IsMatch(this.Server, Resources.Regex_MatchLocalHost, RegexOptions.IgnoreCase)) {
                connect.Username = this.Username;
                connect.Password = this.Password;
                connect.Impersonation = ImpersonationLevel.Impersonate;
                connect.Authentication = AuthenticationLevel.PacketPrivacy;
            }

            //return the new scope
            return new ManagementScope(string.Format(@"\\{0}\{1}", this.Server, path), connect);

        }

        #endregion

        #region Creating And Removing Shares

        //tries to create a new share and return the correct path for it
        public string CreateTemporaryShare(string file) {

            //get the object to create our share
            string config = Path.GetFileName(file);
            string path = file.Substring(0, (file.Length - config.Length) - 1);

            //create the scope to access the server
            ManagementScope scope = null;
            try {
                scope = this._CreateWmiScope(@"root\CIMV2");
                scope.Connect();
            }
            catch {
                throw new InvalidOperationException();
            }


            //send the request
            ManagementObject obj = new ManagementClass(scope, new ManagementPath("Win32_Share"), null);
            try {
                object invoke = obj.InvokeMethod("Create",
                    new object[] {
                    path,
                    WebServerInterface.WEBCONFIG_SHARE_NAME,
                    0,
                    1,
                    "Temp creation for WebConfigEncrypt"
                });

                //if the request displayed an error, notify the user
                if (invoke == null || (invoke is int && (int)invoke != 0)) {
                    throw new ManagementException();
                }

                //return the path
                return string.Format(
                    @"\\{0}\{1}\{2}",
                    this.Server,
                    WebServerInterface.WEBCONFIG_SHARE_NAME,
                    config
                    );

            }
            catch {
                throw new ManagementException();
            }
            finally {
                obj.Dispose();
            }

        }

        //tries to remove a share for WebConfigEncrypt
        public void RemoveTemporaryShare() {

            //create the scope to access the server
            ManagementScope scope = null;
            try {
                scope = this._CreateWmiScope(@"root\CIMV2");
                scope.Connect();
            }
            catch {
                throw new InvalidOperationException();
            }

            //send the request to delete the share
            ManagementObjectSearcher search = null;
            try {
                //remove anything matching - hate deleting stuff... 
                search = new ManagementObjectSearcher(scope, new ObjectQuery(string.Format("select Name from Win32_Share where name = '{0}'", WebServerInterface.WEBCONFIG_SHARE_NAME)));
                foreach (ManagementObject item in search.Get()) {
                    var hold = item;
                    item.InvokeMethod("Delete", null);
                }
            }
            catch {
                throw new ManagementException();
            }
            finally {
                search.Dispose();
            }

        }

        #endregion

    }

}
