﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using WebConfigEncrypt.Interfaces;
using WebConfigEncrypt.Types;
using System.Xml.Linq;
using System.Threading;
using System.Management;
using System.Diagnostics;

namespace WebConfigEncrypt {

    public partial class ConfigListing : Form {

        #region Constructors

        public ConfigListing() {
            InitializeComponent();

            //prepares the server connection
            this.Load += (sender, e) => {

                //initalize our WMI connection
                this._InitWebServerInterface();

                //if no settings were loaded then cancel
                if (this._Interface == null) {
                    Application.Exit();
                    return;
                }

                //download the current server information
                this._RefreshServerList();

                //set the handlers to toggle the views
                this.ServerList.NodeMouseDoubleClick += this._HandleServerListNodeDoubleClicks;
                this.ServerList.NodeMouseClick += this._HandleServerListNodeSingleClicks;

            };

        }

        #endregion

        #region Properties and Fields

        //interface with WMI to handle our server calls
        private WebServerInterface _Interface;

        #endregion

        #region Handling The Form

        //initalizes the Web Interface for this control
        private void _InitWebServerInterface() {

            //get the settings to use
            ConfigureServer config = new ConfigureServer();
            config.Server.Text = "localhost";
            
            DialogResult result = config.ShowDialog();

            //check if this was a cancel
            if (result != DialogResult.OK) { return; }

            //get the settings to use
            this._Interface = new WebServerInterface();
            this._Interface.Server = config.Server.Text ;
            this._Interface.Username = config.Username.Text; ;
            this._Interface.Password = config.Password.Text;
            this._Interface.Domain = config.Domain.Text;
        }

        //update nodes when clicked on
        private void _HandleServerListNodeSingleClicks(object sender, EventArgs e) {
            TreeNodeMouseClickEventArgs press = e as TreeNodeMouseClickEventArgs;
            if (press.Node.Tag is IISDirectory) {
                this._HandleToggleDirectory(press);
            }
        }

        //update nodes when clicked on
        private void _HandleServerListNodeDoubleClicks(object sender, EventArgs e) {
            TreeNodeMouseClickEventArgs press = e as TreeNodeMouseClickEventArgs;
            if (press.Node.Tag is IISDirectory) {
                this._HandleToggleDirectory(press);
            }
            if (press.Node.Tag is WebConfigInformation) {
                this._HandleShowConfig(press);
            }
        }

        //updates the list of our servers
        private void _RefreshServerList() {

            //thread this part separately - painful WinForms threading
            Action work = () => {

                //Get a list of sites, but get their nickname so they
                //can be identified easier
                Dictionary<string, IISDirectory[]> sites = new Dictionary<string, IISDirectory[]>();
                try {
                    sites = this._Interface.GetVirtualDirectories()
                        .GroupBy(o => o.Site)
                        .Select(o => new {
                            Name = this._Interface.GetSiteName(o.Key),
                            Sites = o.ToArray()
                        })
                        .ToDictionary(
                            k => k.Name,
                            v => v.Sites.ToArray()
                                );
                }
                catch (InvalidOperationException) {
                    MessageBox.Show(
                        string.Format("Could not contact server '{0}'", this._Interface.Server),
                        "Nkript - Could not contact server",
                        MessageBoxButtons.OK,
                        MessageBoxIcon.Error
                        );
                }
                catch (ManagementException) {
                    MessageBox.Show(
                        string.Format("Could not query the webserver '{0}' for sites", this._Interface.Server),
                        "Nkript - Could query server",
                        MessageBoxButtons.OK,
                        MessageBoxIcon.Error
                        );
                }
                catch (Exception ex) {
                    MessageBox.Show(
                        string.Format("Unexpected error working with '{0}'\n\n{1}", this._Interface.Server, ex.Message),
                        "Nkript - Unknown Error",
                        MessageBoxButtons.OK,
                        MessageBoxIcon.Error
                        );
                }

                //update the UI - this sure seems like
                //a painful way to have to do this
                Action update = () => {

                    //update the message
                    this.WorkingServer.Text = string.Format("Found {0} sites on '{1}'", sites.Count(), this._Interface.Server);
                    this.Loading.Visible = false;

                    //update the screen
                    foreach (var item in sites.AsEnumerable()) {

                        //builds each node for the user
                        TreeNode node = new TreeNode() {
                            Text = item.Key,
                            Tag = item
                        };
                        this.ServerList.Nodes.Add(node);

                        //add each virtual directory -- even if it has no web.config
                        foreach (IISDirectory directory in item.Value) {
                            TreeNode site = new TreeNode();
                            site.Text = directory.VirtualPath.Equals("/") ? "/ <directory root>" : directory.VirtualPath;
                            site.Tag = directory;
                            node.Nodes.Add(site);
                            site.Nodes.Add("");
                        }

                    }
                };
                this.BeginInvoke(update);
                
            };

            //show the loading
            this.ServerList.Nodes.Clear();
            this.Loading.Visible = true;
            this.WorkingServer.Text = string.Format("Searching '{0}'", this._Interface.Server);
            work.BeginInvoke(null, null);

        }
        
        //Handles showing the information for an IIS directory
        private void _HandleToggleDirectory(TreeNodeMouseClickEventArgs args) {

            //if this is showing some information, update the node
            if (args.Node.IsExpanded) {

                Action work = () => {

                    //get the files found
                    WebConfigInformation[] info = new WebConfigInformation[] { };
                    try {
                        info = this._Interface
                            .FindWebConfigFiles((IISDirectory)args.Node.Tag)
                            .ToArray();
                    }
                    catch (InvalidOperationException) {
                        MessageBox.Show(
                            string.Format("Could not contact server '{0}'", this._Interface.Server),
                            "Nkript - Could not contact server",
                            MessageBoxButtons.OK,
                            MessageBoxIcon.Error
                            );
                    }
                    catch (ManagementException) {
                        MessageBox.Show(
                            string.Format("Could not query the webserver '{0}' for sites", this._Interface.Server),
                            "Nkript - Could query server",
                            MessageBoxButtons.OK,
                            MessageBoxIcon.Error
                            );
                    }
                    catch (Exception ex) {
                        MessageBox.Show(
                            string.Format("Unexpected error working with '{0}'\n\n{1}", this._Interface.Server, ex.Message),
                            "Nkript - Unknown Error",
                            MessageBoxButtons.OK,
                            MessageBoxIcon.Error
                            );
                    }

                    //painful UI updating -- Ideas would be appreciated
                    Action update = () => {
                        bool start = true;
                        foreach (var item in info) {

                            //clear out the 'finding' message
                            if (start) {
                                start = false;
                                args.Node.Nodes.Clear();
                            }

                            //append our node
                            args.Node.Nodes.Add(
                                new TreeNode() {
                                    Text = item.Path,
                                    Tag = item
                                });

                        }

                        //make sure something is there
                        if (info.Count() == 0) {
                            args.Node.Nodes.Clear();
                            args.Node.Nodes.Add("No web.config files were found");
                        }
                    };
                    this.BeginInvoke(update);

                };

                //notify the user what we're up to
                args.Node.Nodes.Clear();
                args.Node.Nodes.Add("Finding web.config files...");
                work.BeginInvoke(null, null);

            }

        }

        //shows a web.config file for a directory
        private void _HandleShowConfig(TreeNodeMouseClickEventArgs args) {

            //check and make sure this file is available
            WebConfigInformation info = args.Node.Tag as WebConfigInformation;

            //read the information for this file
            string target = this._Interface.CreateTemporaryShare(info.Path);
            XDocument file = this._Interface.ReadRemoteFile(target);
            this._Interface.RemoveTemporaryShare();

            //if the file was found - show that the connection was a success
            if (file is XDocument) {
                EncryptSection encrypt = new EncryptSection(file, info, this._Interface);
                encrypt.ShowDialog();
            }
            else {
                MessageBox.Show("Failed to read file " + info.Path);
            }


        }

        #endregion

        #region Menu Options

        private void exitToolStripMenuItem2_Click(object sender, EventArgs e) {
            Application.Exit();
        }

        private void changeServerSettingsToolStripMenuItem1_Click(object sender, EventArgs e) {

            //get the settings to use
            ConfigureServer config = new ConfigureServer();
            config.Password.Text = this._Interface.Password;
            config.ConfirmPassword.Text = this._Interface.Password;
            config.Username.Text = this._Interface.Username;
            config.Domain.Text = this._Interface.Domain;
            config.Server.Text = this._Interface.Server;

            //get the result
            DialogResult result = config.ShowDialog();

            //check if this was a cancel
            if (result != DialogResult.OK) { return; }

            //get the settings to use
            this._Interface = new WebServerInterface();
            this._Interface.Server = config.Server.Text;
            this._Interface.Username = config.Username.Text; ;
            this._Interface.Password = config.Password.Text;
            this._Interface.Domain = config.Domain.Text;

            //update the screen
            this._RefreshServerList();
        }

        private void aboutNkriptToolStripMenuItem1_Click(object sender, EventArgs e) {
            (new About()).ShowDialog();
        }

        #endregion

    }
}
